"""A evaluator based on CPS form and trampolining.



Fixme: add a lot more documentation here about how this all works.

"""

import pogo
from expressions import *
from env import *
import exceptions
import traceback
import types
from parser import parse
from symbol import false


def identity(val):
    """The identity function."""
    return val


def apply_cont(cont, val):
    """Applies a continuation to the given value.
    Actually returns a bouncer value as part of the trampoline.
    """
    return pogo.bounce(cont, val)


"""We need to save Python's apply() function, since we use it later on
to evaluate primitive procedure calls.  For symmetry, I'm also saving
Python's eval() function within evalInUnderlyingPython()."""
evalInUnderlyingPython = eval
applyInUnderlyingPython = apply



######################################################################
## The heart of the interpreter is eval/apply.
##
## Some notes here: I'm using trampolined style to get around Python's
## lack of tail recursion.


def eval(exp, env):
    landing_cont = lambda v: pogo.land(v)
    return pogo.pogo_start(teval, exp, env, landing_cont)


def teval(exp, env, cont):
    """Evaluates an expression 'exp' in an environment 'env'.

    Exercise 4.3 asks us to rewrite this in a more natural
    data-directed manner.  Pychecker, also, doesn't like seeing
    so many 'return' statements in one function.  *grin*
    """
    if isSelfEvaluating(exp):
        return apply_cont(cont, exp)
    if isVariable(exp):
        return apply_cont(cont, lookupVariableValue(exp, env))
    if isQuoted(exp):
        return evalQuoted(exp, env, cont)
    if isAssignment(exp):
        return evalAssignment(exp, env, cont)
    if isDefinition(exp):
        return evalDefinition(exp, env, cont)
    if isLet(exp):
        return teval(letToApplication(exp), env, cont)
    if isIf(exp):
        return evalIf(exp, env, cont)
    if isLambda(exp):
        return apply_cont(cont, makeProcedure(lambdaParameters(exp),
                                              lambdaBody(exp),
                                              env))
    if isBegin(exp):
        return evalSequence(beginActions(exp), env, cont)
    if isCond(exp):
        return teval(condToIf(exp), env, cont)
    if isApplication(exp):
        return evalApplication(exp, env, cont)
    raise SchemeError, "Unknown expression type -- eval " + str(exp)



def apply(procedure, arguments, env, cont):
    """Applies a procedure on a list of arguments."""
    if isPrimitiveProcedure(procedure):
        return applyPrimitiveProcedure(procedure, arguments, env, cont)
    elif isContinuationProcedure(procedure):
        return applyContinuationProcedure(procedure, arguments)
    if isCompoundProcedure(procedure):
        newEnv = extendEnvironment(pairsToPythonList(procedureParameters(procedure)),
                                   pairsToPythonList(arguments),
                                   procedureEnvironment(procedure))
        return evalSequence(procedureBody(procedure), newEnv, cont)
    raise SchemeError, "Unknown procedure type -- apply " + str(procedure)



def applyPrimitiveProcedure(proc, args, env, cont):
    try:
        return applyInUnderlyingPython(primitiveImplementation(proc),
                                       [cont, env, pairsToPythonList(args)])
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e


def applyContinuationProcedure(proc, args):
    try:
        return applyInUnderlyingPython(continuationImplementation(proc),
                                       pairsToPythonList(args))
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e



def evalRands(exps, env, cont):
    """Given a list of expressions, returns a new list containing the
    values of evaluating each on of them.  If the continuation is
    given, then calls cont() on the evaluated operands instead."""
    def c1(head_val):
        def c2(tail_val):
            return apply_cont(cont, schemeCons(head_val, tail_val))
        return evalRands(restOperands(exps), env, c2)
    if isNoOperands(exps):
        return apply_cont(cont, schemeList())
    return teval(firstOperand(exps), env, c1)


def evalIf(exp, env, cont):
    def c(predicate_val):
        if isTrue(predicate_val):
            return teval(ifConsequent(exp), env, cont)
        else:
            return teval(ifAlternative(exp), env, cont)
    return teval(ifPredicate(exp), env, c)


def evalSequence(exps, env, cont):
    def c(val):
        return evalSequence(restExps(exps), env, cont)
    if isLastExp(exps):
        return teval(firstExp(exps), env, cont)
    else:
        return teval(firstExp(exps), env, c)


def evalAssignment(exp, env, cont):
    def c(val):
        setVariableValue(assignmentVariable(exp), val, env)
        return apply_cont(cont, Symbol("ok"))
    return teval(assignmentValue(exp), env, c)


def evalDefinition(exp, env, cont):
    def c(val):
        defineVariable(definitionVariable(exp), val, env)
        return apply_cont(cont, Symbol("ok"))
    return teval(definitionValue(exp), env, c)


def evalApplication(exp, env, cont):
    def c1(operator_val):
        def c2(operands_val):
            return apply(operator_val, operands_val, env, cont)
        return evalRands(operands(exp), env, c2)
    return teval(operator(exp), env, c1)


def evalQuoted(exp, env, cont):
    """Returns a quoted expression, using deepQuotedEval to look for
    UNQUOTE.

    Consequently, quoted elements are properly turned into cons pairs.
    """
    if isQuasiquoted(exp):
        return deepQuotedEval(textOfQuotation(exp), env, 1, cont)
    else:
        return deepQuotedEval(textOfQuotation(exp), env, 0, cont)


def deepQuotedEval(exp, env, withQuasi, cont):
    if not isList(exp):
        return apply_cont(cont, exp)
    if isNull(exp):
        return apply_cont(cont, exp)
    if isUnquoted(exp) and withQuasi:
        return teval(textOfUnquotation(exp), env, cont)
    else:
        def c1(head_val):
            def c2(tail_val):
                return apply_cont(cont, schemeCons(head_val, tail_val))
            return deepQuotedEval(schemeCdr(exp), env, withQuasi, c2)
        return deepQuotedEval(schemeCar(exp), env, withQuasi, c1)



