"""Provides functions to extract portions of the expression tree and
to construct new expressions.  Heavily used in eval.py"""


from pair import *
import types
from symbol import isSymbol, false
import unittest


def isSelfEvaluating(exp):
    """Returns True if the expression is self-evaluating."""
    return isNumber(exp) or isString(exp) or isNull(exp)


def isNumber(x):
    """Returns True if we see a number."""
    return type(x) in (types.IntType, types.FloatType, types.LongType)


def isString(x):
    """Returns True if we see a string."""
    return type(x) == types.StringType



def isVariable(exp):
    """Returns true if the expression looks like a symbol."""
    return isSymbol(exp)


def isTaggedList(exp, tag):
    """Returns true if the expression is tagged with the 'tag'."""
    if isList(exp) and schemeLength(exp) > 0:
        return schemeCar(exp) == tag
    return 0


######################################################################

def isQuoted(exp):
    return (isTaggedList(exp, Symbol("quote")) or
            isTaggedList(exp, Symbol("quasiquote")))

def makeQuoted(exp):
    return schemeList(Symbol("quote"), exp)


def isQuasiquoted(exp):
    return isTaggedList(exp, Symbol("quasiquote"))


def isUnquoted(exp):
    return isTaggedList(exp, Symbol("unquote"))


def textOfQuotation(exp):
    return schemeCadr(exp)


def textOfUnquotation(exp):
    return schemeCadr(exp)


def expandQuasiquotation(exp):
    """Takes a quasiquoted expression and constructs a new quoted
    expression."""
    if not isList(exp):
        return makeQuoted(exp)
    if isUnquoted(exp):
        return textOfUnquotation(exp)
    else:
        return schemeCons(Symbol("list"),
                          listMap(expandQuasiquotation, exp))



######################################################################

def isLet(exp):
    return isTaggedList(exp, Symbol("let"))


def letBindings(exp):
    return schemeCadr(exp)


def letBindingVariables(bindings):
    if isNull(bindings): return schemeList()
    return schemeCons(schemeCar(schemeCar(bindings)),
                      letBindingVariables(schemeCdr(bindings)))

def letBindingValues(bindings):
    if isNull(bindings): return schemeList()
    return schemeCons(schemeCadr(schemeCar(bindings)),
                      letBindingValues(schemeCdr(bindings)))

def letBody(exp):
    return schemeCdr(schemeCdr(exp))

def letToApplication(exp):
    bindings = letBindings(exp)
    return schemeCons(makeLambda(letBindingVariables(bindings),
                                 letBody(exp)),
                      letBindingValues(bindings))

######################################################################


def isAssignment(exp):
    """Assignments have the form (set! <var> <value>)."""
    return isTaggedList(exp, Symbol("set!"))


def assignmentVariable(exp):
    return schemeCadr(exp)


def assignmentValue(exp):
    return schemeCaddr(exp)


######################################################################


def isDefinition(exp):
    return isTaggedList(exp, Symbol("define"))


def definitionVariable(exp):
    if isSymbol(schemeCadr(exp)):
        return schemeCadr(exp)
    return schemeCar(schemeCadr(exp))


def definitionValue(exp):
    if isSymbol(schemeCadr(exp)):
        return schemeCaddr(exp)
    return makeLambda(schemeCdr(schemeCadr(exp)),
                      schemeCddr(exp))


######################################################################

def isLambda(exp):
    return isTaggedList(exp, Symbol("lambda"))


def lambdaParameters(exp):
    return schemeCadr(exp)


def lambdaBody(exp):
    return schemeCddr(exp)


def makeLambda(parameters, body):
    if schemeCount(Symbol("."), parameters) > 1:
        raise SchemeError, \
              "Parameter list can have at most one vararg parameter -- LAMBDA"
    return schemeAppend(schemeList(Symbol("lambda"), parameters),
                        body)


######################################################################


def isIf(exp):
    return isTaggedList(exp, Symbol("if"))


def ifPredicate(exp):
    return schemeCadr(exp)


def ifConsequent(exp):
    return schemeCaddr(exp)


def ifAlternative(exp):
    if not isNull(schemeCdddr(exp)):
        return schemeCadddr(exp)
    return false


def makeIf(predicate, consequent, alternative):
    return schemeList(Symbol("if"), predicate, consequent, alternative)


######################################################################

def isBegin(exp):
    return isTaggedList(exp, Symbol("begin"))


def beginActions(exp):
    return schemeCdr(exp)


def isLastExp(seq):
    return isNull(schemeCdr(seq))


def firstExp(seq):
    return schemeCar(seq)


def restExps(seq):
    return schemeCdr(seq)


def sequenceToExp(seq):
    if isNull(seq): return seq
    if isLastExp(seq): return firstExp(seq)
    return makeBegin(seq)


def makeBegin(seq):
    return schemeCons(Symbol("begin"), seq)


######################################################################


def isApplication(exp):
    return isList(exp)


def operator(exp):
    if schemeLength(exp) == 0:
        raise SchemeError, \
              "No operator given for procedure application -- OPERATOR"
    return schemeCar(exp)


def operands(exp):
    return schemeCdr(exp)


def isNoOperands(ops):
    return isNull(ops)


def firstOperand(ops):
    return schemeCar(ops)


def restOperands(ops):
    return schemeCdr(ops)


######################################################################


def isCond(exp):
    return isTaggedList(exp, Symbol("cond"))


def condClauses(exp):
    return schemeCdr(exp)


def isCondElseClause(clause):
    return condPredicate(clause) == Symbol("else")


def condPredicate(clause):
    return schemeCar(clause)


def condActions(clause):
    return schemeCdr(clause)


def condToIf(exp):
    return expandClauses(condClauses(exp))


def expandClauses(clauses):
    if isNull(clauses):
        return false
    first = schemeCar(clauses)
    rest = schemeCdr(clauses)
    if isCondElseClause(first):
        if isNull(rest):
            return sequenceToExp(condActions(first))
        raise SchemeError, "else clause isn't last -- condToIf" + clauses
    return makeIf(condPredicate(first),
                  sequenceToExp(condActions(first)),
                  expandClauses(rest))

def isTrue(x):
    return (x != false)


def isFalse(x):
    return (x == false)

                                
######################################################################


def makeProcedure(parameters, body, env):
    return schemeList(Symbol("procedure"), parameters, body, env)

def isCompoundProcedure(p):
    return isTaggedList(p, Symbol("procedure"))

def procedureParameters(p):
    return schemeCadr(p)

def procedureBody(p):
    return schemeCaddr(p)

def procedureEnvironment(p):
    return schemeCadddr(p)



######################################################################
"""Here's where we define the primitives of Scheme."""
def isPrimitiveProcedure(proc):
    return isTaggedList(proc, Symbol("primitive"))


def primitiveImplementation(proc):
    return schemeCadr(proc)


def makePrimitiveProcedure(proc):
    return schemeList(Symbol("primitive"), proc)



## We have to tag a continuation slightly different: applying a
## continuation abandons the current computation.
def isContinuationProcedure(proc):
    return isTaggedList(proc, Symbol("continuation"))


def continuationImplementation(proc):
    return schemeCadr(proc)


def makeContinuationProcedure(proc):
    """Procedure is defined to be a function that takes a single 'val'
    argument."""
    return schemeList(Symbol("continuation"), proc)



## Fixme: not tail recursive!
def expressionToString(expr):
    """Given a Scheme expression, returns a string that tries to nicely
    return it as a string."""
    if not isPair(expr):
        if type(expr) is str:
            return repr(expr) ## get quotes around strings.
        return str(expr)
    if isPrimitiveProcedure(expr):
        return ('(%s %s <procedure-env>)') % (
            Symbol('primitive-procedure'),
            primitiveImplementation(expr))
    if isCompoundProcedure(expr):
        return '(%s %s %s <procedure-env>)' % \
               (Symbol('compound-procedure'),
                expressionToString(procedureParameters(expr)),
                expressionToString(procedureBody(expr)))
    pieces = []
    while not isNull(expr):
        pieces.append(expressionToString(schemeCar(expr)))
        if isDottedPair(expr):
            pieces.append(".")
            pieces.append(expressionToString(schemeCdr(expr)))
            break
        expr = schemeCdr(expr)
    return '(' + ' '.join(pieces) + ')'
        
