import unittest

import sys

import analyzer
import parser
import builtins
import pogo
import test_scheme
from env import THE_EMPTY_ENVIRONMENT, extendEnvironment, defineVariable
from symbol import Symbol
from pair import schemeList

from error import SchemeError
import analyzer_builtins


class AnalyzerEvaluatorMixin(object):
    def pe(self, s):
        """Parse and evaluate."""
        return analyzer.eval(parser.parse(s),
                             self.env)


    
class AnalyzerTests(AnalyzerEvaluatorMixin, unittest.TestCase):
    def setUp(self):
        ## We set up a VERY minimal environment here for some tests.
        ## We also set the recursion limit to something dreadful to see that
        ## call/cc is doing the right thing.
        self.env = extendEnvironment([Symbol('pi')],
                                     [3.1415926],
                                     THE_EMPTY_ENVIRONMENT)
        defineVariable(Symbol("#t"), Symbol("#t"), self.env)
        defineVariable(Symbol("#f"), Symbol("#f"), self.env)
        sys.setrecursionlimit(100)


    def tearDown(self):
        self.env = None


    def testSelfEvaluating(self):
        self.assertEquals(42, self.pe("42"))
        self.assertEquals(schemeList(), self.pe("()"))


    def testVariable(self):
        self.assertEquals(3.1415926, self.pe("pi"))
        self.assertRaises(SchemeError, self.pe, "nonexistantvalue")


    def testQuotation(self):
        self.assertEquals(Symbol("foobar"), self.pe("'foobar"))
        self.assertEquals(Symbol("foobar"), self.pe("'foobar"))
        self.assertEquals(schemeList(1, 2, 3), self.pe("'(1 2 3)"))


    def testAssignment(self):
        self.assertEquals(Symbol("ok"),
                          self.pe("(set! pi 'three-point-one-four)"))
        self.assertEquals(Symbol("three-point-one-four"),
                          self.pe("pi"))
        self.assertRaises(SchemeError, self.pe, "(set! nonexistantvalue 42)")



    def testDefinition(self):
        self.assertEquals(Symbol("ok"),
                          self.pe("(define name 'danny)"))
        self.assertEquals(Symbol("danny"), self.pe("name"))


    def testIf(self):
        self.assertEquals(Symbol("ok"), self.pe("(if #t 'ok 'not-ok)"))
        self.assertEquals(Symbol("ok"), self.pe("(if #f 'not-ok 'ok)"))


    def testLambda(self):
        ## Can't test much yet without application.
        self.pe("(lambda (x) 'hello 'world)")

        
    def testBegin(self):
        self.assertEquals(Symbol("hello"),
                          self.pe("(begin 'hello)"))
        self.assertEquals(Symbol("world"),
                          self.pe("(begin 'hello '(this is a test) 'world)"))
        self.assertRaises(SchemeError, self.pe, "(begin)")

        
    def testSimpleApplication(self):
        self.assertEquals(Symbol("hello"),
                          self.pe("((lambda (x) x) 'hello)"))
        self.assertEquals(Symbol("hi"),
                          self.pe("((lambda (x y) y) '(this is a test) 'hi)"))





class AnalyzerTestsWithBuiltins(AnalyzerEvaluatorMixin,
                                test_scheme.BasicSchemeTests):
    def setUp(self):
        self.env = analyzer_builtins.setupEnvironment()
        sys.setrecursionlimit(100)

    def tearDown(self):
        self.env = None


class OtherAnalyzerTests(AnalyzerEvaluatorMixin,
                         test_scheme.ExtendedSchemeTests):
    def setUp(self):
        self.env = analyzer_builtins.setupEnvironment()
        sys.setrecursionlimit(100)

    def tearDown(self):
        self.env = None


if __name__ == '__main__':
    unittest.main()
